@@ -8,34 +8,91 @@ Functions for grouping values and rows.
 import numpy as np
 
 from . import util
+
 from .constants import log, tol
 
 try:
-    from scipy.spatial import cKDTree as KDTree
-except ImportError:
-    log.warning('Scipy unavailable')
+    from scipy.spatial import cKDTree
+except BaseException as E:
+    # wrapping just ImportError fails in some cases
+    # will raise the error when someone tries to use KDtree
+    from . import exceptions
+    cKDTree = exceptions.closure(E)
 
 
-def merge_vertices(mesh, distance=None):
+def merge_vertices(mesh,
+                   merge_tex=False,
+                   merge_norm=False,
+                   digits_vertex=None,
+                   digits_norm=2,
+                   digits_uv=4,
+                   **kwargs):
     """
-    Removes duplicate vertices, based on integer hashes of
+    Removes duplicate vertices. By default,  based on integer hashes of
     each row.
 
     Parameters
     -------------
     mesh : Trimesh object
       Mesh to merge vertices on
-    distance : float or None
-      If not specified uses tol.merge
+    merge_tex : bool
+      If True textured meshes with UV coordinates will
+      have vertices merged regardless of UV coordinates
+    merge_norm : bool
+      If True, meshes with vertex normals will have
+      vertices merged ignoring different normals
+    digits_vertex : None or int
+      Number of digits to consider for vertex position
+    digits_norm : int
+      Number of digits to consider for unit normals
+    digits_uv : int
+      Number of digits to consider for UV coordinates
     """
-    if distance is not None:
-        digits = util.decimal_to_digits(distance)
+    # use tol.merge if digit precision not passed
+    if not isinstance(digits_vertex, int):
+        digits_vertex = util.decimal_to_digits(tol.merge)
+
+    # if we have a ton of unreferenced vertices it will
+    # make the unique_rows call super slow so cull first
+    if hasattr(mesh, 'faces') and len(mesh.faces) > 0:
+        referenced = np.zeros(len(mesh.vertices), dtype=np.bool)
+        referenced[mesh.faces] = True
     else:
-        digits = None
-    # unique rows
-    unique, inverse = unique_rows(mesh.vertices,
-                                  digits=digits)
-    mesh.update_vertices(unique, inverse)
+        # this is used for geometry without faces
+        referenced = np.ones(len(mesh.vertices), dtype=np.bool)
+
+    # collect vertex attributes into sequence we can stack
+    stacked = [mesh.vertices * (10 ** digits_vertex)]
+
+    # UV texture visuals require us to update the
+    # vertices and normals differently
+    if (not merge_tex and
+        mesh.visual.defined and
+        mesh.visual.kind == 'texture' and
+        mesh.visual.uv is not None and
+            len(mesh.visual.uv) == len(mesh.vertices)):
+        # get an array with vertices and UV coordinates
+        # converted to integers at requested precision
+        stacked.append(mesh.visual.uv * (10 ** digits_uv))
+
+    # check to see if we have vertex normals
+    normals = mesh._cache['vertex_normals']
+    if not merge_norm and np.shape(normals) == mesh.vertices.shape:
+        stacked.append(normals * (10 ** digits_norm))
+
+    # stack collected vertex properties and round to integer
+    stacked = np.column_stack(stacked).round().astype(np.int64)
+
+    # check unique rows of referenced vertices
+    u, i = unique_rows(stacked[referenced])
+
+    # construct an inverse using the subset
+    inverse = np.zeros(len(mesh.vertices), dtype=np.int64)
+    inverse[referenced] = i
+    # get the vertex mask
+    mask = np.nonzero(referenced)[0][u]
+    # run the update including normals and UV coordinates
+    mesh.update_vertices(mask=mask, inverse=inverse)
 
 
 def group(values, min_len=0, max_len=np.inf):
@@ -44,16 +101,20 @@ def group(values, min_len=0, max_len=np.inf):
 
     Parameters
     ----------
-    values:     1D array
-    min_len:    int, the shortest group allowed
-                All groups will have len >= min_length
-    max_len:    int, the longest group allowed
-                All groups will have len <= max_length
+    values : (n,) int
+      Values to group
+    min_len : int
+      The shortest group allowed
+      All groups will have len >= min_length
+    max_len : int
+      The longest group allowed
+      All groups will have len <= max_length
 
     Returns
     ----------
-    groups: sequence of indices to form groups
-            IE [0,1,0,1] returns [[0,2], [1,3]]
+    groups : sequence
+      Contains indices to form groups
+      IE [0,1,0,1] returns [[0,2], [1,3]]
     """
     original = np.asanyarray(values)
 
@@ -91,14 +152,17 @@ def hashable_rows(data, digits=None):
 
     Parameters
     ---------
-    data:    (n,m) input array
-    digits:  how many digits to add to hash, if data is floating point
-             If none, TOL_MERGE will be turned into a digit count and used.
+    data : (n, m) array
+      Input data
+    digits : int or None
+      How many digits to add to hash if data is floating point
+      If None, tol.merge will be used
 
     Returns
     ---------
-    hashable:  (n) length array of custom data which can be sorted
-                or used as hash keys
+    hashable : (n,) array
+      Custom data type which can be sorted
+      or used as hash keys
     """
     # if there is no data return immediately
     if len(data) == 0:
@@ -143,13 +207,17 @@ def float_to_int(data, digits=None, dtype=np.int32):
 
     Parameters
     -------------
-    data:   (n, d) float, int, or bool data
-    digits: float/int precision for float conversion
-    dtype:  numpy dtype for result
+    data :  (n, d) float, int, or bool
+      Input data
+    digits : float or int
+      Precision for float conversion
+    dtype : numpy.dtype
+      What datatype should result be returned as
 
     Returns
     -------------
-    as_int: data, as integers
+    as_int : (n, d) int
+      Data as integers
     """
     # convert to any numpy array
     data = np.asanyarray(data)
@@ -165,7 +233,7 @@ def float_to_int(data, digits=None, dtype=np.int32):
     elif isinstance(digits, float) or isinstance(digits, np.float):
         digits = util.decimal_to_digits(digits)
     elif not (isinstance(digits, int) or isinstance(digits, np.integer)):
-        log.warn('Digits were passed as %s!', digits.__class__.__name__)
+        log.warning('Digits were passed as %s!', digits.__class__.__name__)
         raise ValueError('Digits must be None, int, or float!')
 
     # data is float so convert to large integers
@@ -201,6 +269,75 @@ def unique_ordered(data):
     return result
 
 
+def unique_bincount(values,
+                    minlength=0,
+                    return_inverse=False,
+                    return_counts=False):
+    """
+    For arrays of integers find unique values using bin counting.
+    Roughly 10x faster for correct input than np.unique
+
+    Parameters
+    --------------
+    values : (n,) int
+      Values to find unique members of
+    minlength : int
+      Maximum value that will occur in values (values.max())
+    return_inverse : bool
+      If True, return an inverse such that unique[inverse] == values
+    return_counts : bool
+      If True, also return the number of times each
+      unique item appears in values
+
+    Returns
+    ------------
+    unique : (m,) int
+      Unique values in original array
+    inverse : (n,) int, optional
+      An array such that unique[inverse] == values
+      Only returned if return_inverse is True
+    counts : (m,) int, optional
+      An array holding the counts of each unique item in values
+      Only returned if return_counts is True
+    """
+    values = np.asanyarray(values)
+    if len(values.shape) != 1 or values.dtype.kind != 'i':
+        raise ValueError('input must be 1D integers!')
+
+    try:
+        # count the number of occurrences of each value
+        counts = np.bincount(values, minlength=minlength)
+    except TypeError:
+        # casting failed on 32 bit windows
+        log.warning('casting failed, falling back!')
+        # fall back to numpy unique
+        return np.unique(values,
+                         return_inverse=return_inverse,
+                         return_counts=return_counts)
+
+    # which bins are occupied at all
+    # counts are integers so this works
+    unique_bin = counts.astype(np.bool)
+
+    # which values are unique
+    # indexes correspond to original values
+    unique = np.where(unique_bin)[0]
+    ret = (unique,)
+
+    if return_inverse:
+        # find the inverse to reconstruct original
+        inverse = (np.cumsum(unique_bin) - 1)[values]
+        ret += (inverse,)
+
+    if return_counts:
+        unique_counts = counts[unique]
+        ret += (unique_counts,)
+
+    if len(ret) == 1:
+        return ret[0]
+    return ret
+
+
 def merge_runs(data, digits=None):
     """
     Merge duplicate sequential values. This differs from unique_ordered
@@ -269,37 +406,47 @@ def unique_rows(data, digits=None):
 
     Parameters
     ---------
-    data: (n,m) set of floating point data
-    digits: how many digits to consider for the purposes of uniqueness
+    data : (n, m) array
+      Floating point data
+    digits : int or None
+      How many digits to consider
 
     Returns
     --------
-    unique:  (j) array, index in data which is a unique row
-    inverse: (n) length array to reconstruct original
-                 example: unique[inverse] == data
+    unique :  (j,) int
+      Index in data which is a unique row
+    inverse : (n,) int
+      Array to reconstruct original
+      Example: unique[inverse] == data
     """
     hashes = hashable_rows(data, digits=digits)
-    garbage, unique, inverse = np.unique(hashes,
-                                         return_index=True,
-                                         return_inverse=True)
+    garbage, unique, inverse = np.unique(
+        hashes,
+        return_index=True,
+        return_inverse=True)
     return unique, inverse
 
 
 def unique_value_in_row(data, unique=None):
     """
-    For a 2D array of integers find the position of a value in each
-    row which only occurs once. If there are more than one value per
-    row which occur once, the last one is returned.
+    For a 2D array of integers find the position of a
+    value in each row which only occurs once.
+
+    If there are more than one value per row which
+    occur once, the last one is returned.
 
     Parameters
     ----------
-    data:   (n,d) int
-    unique: (m) int, list of unique values contained in data.
-             speedup purposes only, generated from np.unique if not passed
+    data :   (n, d) int
+      Data to check values
+    unique : (m,) int
+      List of unique values contained in data.
+      Generated from np.unique if not passed
 
     Returns
     ---------
-    result: (n,d) bool, with one or zero True values per row.
+    result : (n, d) bool
+      With one or zero True values per row.
 
 
     Examples
@@ -340,33 +487,35 @@ def group_rows(data, require_count=None, digits=None):
     Returns index groups of duplicate rows, for example:
     [[1,2], [3,4], [1,2]] will return [[0,2], [1]]
 
+
+    Note that using require_count allows numpy advanced
+    indexing to be used in place of looping and
+    checking hashes and is ~10x faster.
+
+
     Parameters
     ----------
-    data:          (n,m) array
-    require_count: only returns groups of a specified length, eg:
-                   require_count =  2
-                   [[1,2], [3,4], [1,2]] will return [[0,2]]
-
-                   Note that using require_count allows numpy advanced indexing
-                   to be used in place of looping and checking hashes, and as a
-                   consequence is ~10x faster.
-
-    digits:        If data is floating point, how many decimals to look at.
-                   If this is None, the value in TOL_MERGE will be turned into a
-                   digit count and used.
+    data : (n, m) array
+      Data to group
+    require_count : None or int
+      Only return groups of a specified length, eg:
+      require_count =  2
+      [[1,2], [3,4], [1,2]] will return [[0,2]]
+    digits : None or int
+    If data is floating point how many decimals
+    to consider, or calculated from tol.merge
 
     Returns
     ----------
-    groups:        List or sequence of indices from data indicating identical rows.
-                   If require_count != None, shape will be (j, require_count)
-                   If require_count is None, shape will be irregular (AKA a sequence)
+    groups : sequence (*,) int
+      Indices from in indicating identical rows.
     """
 
     def group_dict():
         """
         Simple hash table based grouping.
-        The loop and appends make this rather slow on very large arrays,
-        but it works on irregular groups.
+        The loop and appends make this rather slow on
+        large arrays but it works on irregular groups.
         """
         observed = dict()
         hashable = hashable_rows(data, digits=digits)
@@ -501,7 +650,7 @@ def group_distance(values, distance):
 
     consumed = np.zeros(len(values),
                         dtype=np.bool)
-    tree = KDTree(values)
+    tree = cKDTree(values)
 
     # (n, d) set of values that are unique
     unique = []
@@ -537,7 +686,7 @@ def clusters(points, radius):
 
     """
     from . import graph
-    tree = KDTree(points)
+    tree = cKDTree(points)
 
     # some versions return pairs as a set of tuples
     pairs = tree.query_pairs(r=radius, output_type='ndarray')
@@ -550,50 +699,106 @@ def clusters(points, radius):
 def blocks(data,
            min_len=2,
            max_len=np.inf,
+           wrap=False,
            digits=None,
            only_nonzero=False):
     """
-    Given an array, find the indices of contiguous blocks
+    Find the indices in an array of contiguous blocks
     of equal values.
 
     Parameters
-    ---------
-    data:    (n) array
-    min_len: int, the minimum length group to be returned
-    max_len: int, the maximum length group to be retuurned
-    digits:  if dealing with floats, how many digits to use
-    only_nonzero: bool, only return blocks of non- zero values
+    ------------
+    data : (n,) array
+      Data to find blocks on
+    min_len : int
+      The minimum length group to be returned
+    max_len : int
+      The maximum length group to be retuurned
+    wrap : bool
+      Combine blocks on both ends of 1D array
+    digits : None or int
+      If dealing with floats how many digits to consider
+    only_nonzero : bool
+      Only return blocks of non- zero values
 
     Returns
     ---------
-    blocks: (m) sequence of indices referencing data
+    blocks : (m) sequence of (*,) int
+      Indices referencing data
     """
     data = float_to_int(data, digits=digits)
 
-    # find the inflection points, or locations where the array turns
-    # from True to False.
+    # find the inflection points
+    # AKA locations where the array goes from True to False.
     infl = np.concatenate(([0],
                            np.nonzero(np.diff(data))[0] + 1,
                            [len(data)]))
     infl_len = np.diff(infl)
+    # check the length of each group
     infl_ok = np.logical_and(infl_len >= min_len,
                              infl_len <= max_len)
 
     if only_nonzero:
-        # check to make sure the values of each contiguous block are True,
-        # by checking the first value of each block
-        infl_ok = np.logical_and(infl_ok,
-                                 data[infl[:-1]])
+        # check to make sure the values of each contiguous block
+        # are True by checking the first value of each block
+        infl_ok = np.logical_and(
+            infl_ok, data[infl[:-1]])
 
     # inflate start/end indexes into full ranges of values
     blocks = [np.arange(infl[i], infl[i + 1])
               for i, ok in enumerate(infl_ok) if ok]
+
+    if wrap:
+        # wrap only matters if first and last points are the same
+        if data[0] != data[-1]:
+            return blocks
+        # if we are only grouping nonzero things and
+        # the first and last point are zero we can exit
+        if only_nonzero and not bool(data[0]):
+            return blocks
+
+        # so now first point equals last point, so the cases are:
+        # - first and last point are in a block: combine two blocks
+        # - first OR last point are in block: add other point to block
+        # - neither are in a block: check if combined is eligible block
+
+        # first point is in a block
+        first = len(blocks) > 0 and blocks[0][0] == 0
+        # last point is in a block
+        last = len(blocks) > 0 and blocks[-1][-1] == (len(data) - 1)
+
+        # CASE: first and last point are BOTH in block: combine blocks
+        if first and last:
+            blocks[0] = np.append(blocks[-1], blocks[0])
+            blocks.pop()
+        else:
+            # combined length
+            combined = infl_len[0] + infl_len[-1]
+            # exit if lengths aren't OK
+            if combined < min_len or combined > max_len:
+                return blocks
+            # new block combines both ends
+            new_block = np.append(np.arange(infl[-2], infl[-1]),
+                                  np.arange(infl[0], infl[1]))
+            # we are in a first OR last situation now
+            if first:
+                # first was already in a block so replace it with combined
+                blocks[0] = new_block
+            elif last:
+                # last was already in a block so replace with superset
+                blocks[-1] = new_block
+            else:
+                # both are false
+                # combined length generated new block
+                blocks.append(new_block)
+
     return blocks
 
 
 def group_min(groups, data):
     """
-    Given a list of groups, find the minimum element of data within each group
+    Given a list of groups find the minimum element of data
+    within each group
 
     Parameters
     -----------
